/*
 * Copyright 2001-2023 Geert Bevin (gbevin[remove] at uwyn dot com)
 * Licensed under the Apache License, Version 2.0 (the "License")
 */
package rife.bld;

import rife.bld.help.*;
import rife.bld.operations.*;
import rife.bld.publish.PublishArtifact;

import java.io.File;
import java.util.List;
import java.util.Objects;

/**
 * Provides the configuration and commands of a Java web project for the
 * build system.
 *
 * @author Geert Bevin (gbevin[remove] at uwyn dot com)
 * @since 1.5
 */
public class WebProject extends Project {
    /**
     * The main war archive creation.
     *
     * @see #warFileName()
     * @since 1.5.0
     */
    protected String warFileName = null;
    /**
     * The main webapp directory.
     *
     * @see #srcMainWebappDirectory()
     * @since 1.5
     */
    protected File srcMainWebappDirectory = null;

    /*
     * Standard build commands
     */

    private final UberJarOperation uberJarOperation_ = new UberJarOperation();
    private final WarOperation warOperation_ = new WarOperation();
    private final PublishOperation publishOperation_ = new PublishOperation();

    /**
     * Standard build command, creates an UberJar archive for the web project.
     *
     * @since 1.5
     */
    @Override
    public void uberjar()
    throws Exception {
        jar();
        uberJarOperation_.executeOnce(o -> {
            o.fromProject(this);
            o.sourceDirectories(List.of(new NamedFile("webapp", srcMainWebappDirectory())));
            o.jarSourceFiles().addAll(standaloneClasspathJars());

        });
    }

    /**
     * Standard build command, creates a war archive for the web project.
     *
     * @since 1.5
     */
    @BuildCommand(help = WarHelp.class)
    public void war()
    throws Exception {
        jar();
        warOperation_.executeOnce(o -> o.fromProject(this));
    }

    /**
     * Standard publish-web command, uploads artifacts to the publication repository.
     *
     * @since 1.5.7
     */
    @BuildCommand(help = PublishWebHelp.class)
    public void publish()
    throws Exception {
        jar();
        jarSources();
        jarJavadoc();
        uberjar();
        war();
        publishOperation_.executeOnce(o -> {
            o.fromProject(this);
            o.artifacts().add(new PublishArtifact(new File(buildDistDirectory(), uberJarFileName()), "uber", "jar"));
            o.artifacts().add(new PublishArtifact(new File(buildDistDirectory(), warFileName()), "", "war"));
        });
    }

    /*
     * Project directories
     */

    @Override
    public File libStandaloneDirectory() {
        return Objects.requireNonNullElseGet(libStandaloneDirectory, () -> new File(libDirectory(), "standalone"));
    }

    /**
     * Returns the project main webapp directory.
     * Defaults to {@code "webapp"} relative to {@link #srcMainDirectory()}.
     *
     * @since 1.5
     */
    public File srcMainWebappDirectory() {
        return Objects.requireNonNullElseGet(srcMainWebappDirectory, () -> new File(srcMainDirectory(), "webapp"));
    }

    /*
     * Project options
     */

    /**
     * Returns filename to use for the main war archive creation.
     * By default, appends the version and the {@code war} extension to the {@link #archiveBaseName()}.
     *
     * @since 1.5.0
     */
    public String warFileName() {
        return Objects.requireNonNullElseGet(warFileName, () -> archiveBaseName() + "-" + version() + ".war");
    }
}
